.header {
  width: 100%;
  background-color: var(--color-nav-background);
  box-shadow: inset 0 -1px 0 0 var(--color-keyline-default);
  padding: 8px 16px;
  display: flex;
  flex-direction: row;
  align-items: center;
}

.header a {
  line-height: 0;
}

.svg {
  height: 48px;
  width: 168px;
}
